/*
 * Copyright (c) 2007-2010 Sonatype, Inc.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package org.sonatype.plugins.sisu.assembler;

import java.io.File;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.Set;

import org.apache.maven.model.Repository;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.MavenProjectHelper;
import org.codehaus.plexus.util.FileUtils;
import org.sonatype.eclipse.bridge.EclipseInstance;
import org.sonatype.p2.bridge.ArtifactRepository;
import org.sonatype.p2.bridge.MetadataRepository;
import org.sonatype.p2.bridge.P2Director;
import org.sonatype.p2.bridge.P2ProfileRegistry;
import org.sonatype.sisu.assembler.Assembler;
import org.sonatype.sisu.assembler.Fragment;
import org.sonatype.sisu.assembler.Services;

/**
 * @goal assemble
 * @phase package
 * @requiresDependencyResolution runtime
 */
public class AssembleMojo
    extends WithinEclipseRunningProjectBasedMojo
{

    /**
     * @parameter expression="${project.build.directory}/p2/p2-install-folder"
     * @required
     */
    private File installDirectory;

    /**
     * @parameter expression="${project.build.directory}/p2/p2-update-sites"
     * @required
     */
    private File updateSitesDirectory;

    /**
     * @parameter expression="${project.build.directory}/p2/recipe.xml"
     * @readOnly
     */
    private File generatedRecipeFile;

    /**
     * @parameter
     */
    private List<Recipe> recipes;

    /**
     * @parameter
     */
    private List<Fragment> fragments;

    /**
     * @parameter
     */
    private Map<String, String> properties;

    /**
     * @parameter default-value="false"
     */
    private boolean useProjectProperties;

    /**
     * @component
     */
    private MavenProjectHelper projectHelper;

    /**
     * @component
     */
    private Assembler assembler;

    @Override
    protected void doWithEclipse( final EclipseInstance eclipse )
        throws MojoExecutionException
    {
        Utils.failIfTouchpointClassesAvailableInClasspath();
        final Map<String, String> assembleProps = new HashMap<String, String>();
        if ( useProjectProperties )
        {
            if ( project.getProperties() != null )
            {
                for ( final Map.Entry<Object, Object> entry : project.getProperties().entrySet() )
                {
                    if ( entry.getValue() != null )
                    {
                        assembleProps.put( entry.getKey().toString(), entry.getValue().toString() );
                    }
                }
            }
        }
        else if ( properties != null )
        {
            assembleProps.putAll( properties );
        }

        resolveRecipesVersionFromDependencyManagement();
        resolveFragmentsVersionFromDependencyManagement();

        try
        {
            FileUtils.deleteDirectory( installDirectory );
            FileUtils.deleteDirectory( updateSitesDirectory );

            final Properties systemProperties = new Properties();
            systemProperties.putAll( System.getProperties() );
            systemProperties.putAll( project.getProperties() );

            assembler.assemble(
                recipes,
                fragments,
                installDirectory,
                getRepositoriesWithP2Layout(),
                assembleProps,
                systemProperties,
                updateSitesDirectory,
                generatedRecipeFile,
                new Services( this,
                    eclipse.getService( P2Director.class ), eclipse.getService( P2ProfileRegistry.class ),
                    eclipse.getService( ArtifactRepository.class ), eclipse.getService( MetadataRepository.class ) ) );

            projectHelper.attachArtifact( project, "xml", "recipe", generatedRecipeFile );
        }
        catch ( final Exception e )
        {
            throw new MojoExecutionException( String.format( "Cannot assemble %s", project ), e );
        }
    }

    private void resolveRecipesVersionFromDependencyManagement()
        throws MojoExecutionException
    {
        if ( recipes != null )
        {
            for ( final Recipe recipe : recipes )
            {
                if ( recipe.getVersion() == null )
                {
                    recipe.setVersion( resolveVersionFromDependencyManagement( recipe.getGroupId(),
                        recipe.getArtifactId(), null, null ) );
                }
            }
        }
    }

    private void resolveFragmentsVersionFromDependencyManagement()
        throws MojoExecutionException
    {
        if ( fragments != null )
        {
            for ( final Fragment fragment : fragments )
            {
                if ( fragment.getVersion() == null )
                {
                    fragment.setVersion( resolveVersionFromDependencyManagement( fragment.getGroupId(),
                        fragment.getArtifactId(), null, null ) );
                }
            }
        }
    }

    private Collection<String> getRepositoriesWithP2Layout()
    {
        final Set<String> repos = new HashSet<String>();
        final List<Repository> repositories = project.getRepositories();
        if ( repositories != null )
        {
            for ( final Repository repository : repositories )
            {
                if ( "p2".equalsIgnoreCase( repository.getLayout() ) )
                {
                    repos.add( repository.getUrl() );
                }
            }
        }
        return repos;
    }

}
