/*
 * Copyright (c) 2007-2010 Sonatype, Inc.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package org.sonatype.plugins.sisu.assembler;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.apache.maven.RepositoryUtils;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.project.DefaultDependencyResolutionRequest;
import org.apache.maven.project.DependencyResolutionException;
import org.apache.maven.project.DependencyResolutionResult;
import org.apache.maven.project.MavenProjectHelper;
import org.apache.maven.project.ProjectDependenciesResolver;
import org.codehaus.plexus.util.DirectoryScanner;
import org.codehaus.plexus.util.FileUtils;
import org.codehaus.plexus.util.IOUtil;
import org.sonatype.aether.RepositorySystemSession;
import org.sonatype.aether.artifact.Artifact;
import org.sonatype.aether.graph.Dependency;
import org.sonatype.aether.graph.DependencyNode;
import org.sonatype.aether.resolution.ArtifactResolutionException;
import org.sonatype.aether.util.DefaultRepositorySystemSession;
import org.sonatype.aether.util.artifact.JavaScopes;
import org.sonatype.aether.util.graph.DefaultDependencyNode;
import org.sonatype.aether.util.graph.transformer.ChainedDependencyGraphTransformer;
import org.sonatype.eclipse.bridge.EclipseInstance;
import org.sonatype.p2.bridge.ArtifactRepository;
import org.sonatype.p2.bridge.BundlePublisher;
import org.sonatype.p2.bridge.MetadataRepository;
import org.sonatype.sisu.assembler.AdvicesParser;
import org.sonatype.sisu.assembler.Generator;
import org.sonatype.sisu.assembler.GeneratorResult;
import org.sonatype.sisu.assembler.InstallableUnitsGenerator;
import org.sonatype.sisu.assembler.VersionPolicy;
import org.sonatype.sisu.assembler.advice.Advice;
import org.sonatype.sisu.assembler.advice.GenerateSettings;
import org.sonatype.sisu.assembler.advice.InstallableUnit;
import org.sonatype.sisu.assembler.advice.OnOffSwitch;

/**
 * @goal generate
 * @phase package
 * @requiresDependencyResolution runtime
 */
public class GenerateMojo
    extends WithinEclipseRunningProjectBasedMojo
{

    /**
     * @parameter expression="${project.groupId}:${project.artifactId}"
     * @required
     */
    private String id;

    /**
     * @parameter expression="${project.groupId}:${project.artifactId}"
     * @readonly
     * @required
     */
    private String projectId;

    /**
     * @parameter expression="${sisu.assembler.tags}"
     */
    private String tags;

    /**
     * @parameter expression="src/main/advices"
     * @required
     */
    private File advicesDirectory;

    /**
     * @parameter
     */
    private List<AdviceArtifact> adviceArtifacts;

    /**
     * @parameter expression="${project.build.directory}/p2/p2-update-site"
     * @required
     */
    protected File updateSiteDirectory;

    /**
     * @parameter expression="${sisu.assembler.attachUpdateSiteXmls}" default-value="true"
     */
    private boolean attachUpdateSiteXmls;

    /**
     * @parameter expression="${sisu.assembler.useDefaultMetaRequirements}" default-value="true"
     */
    private boolean useDefaultMetaRequirements;

    /**
     * @parameter expression="${sisu.assembler.useTransitiveDependencies}" default-value="true"
     */
    private boolean useTransitiveDependencies;

    /**
     * @parameter
     */
    private ClassifierAndTypeSpec alternativeArtifact;

    /**
     * @parameter
     */
    private Set<MetaRequirement> metaRequirements;

    /**
     * @parameter
     */
    private VersionPolicy versionPolicy;

    /**
     * @parameter expression="${sisu.assembler.resolveUpdateSiteArtifacts}" default-value="false"
     */
    private boolean resolveUpdateSiteArtifacts;

    /**
     * @parameter
     */
    private List<String> includeScopes;

    /**
     * @parameter
     */
    private List<String> excludeScopes = Arrays.asList( JavaScopes.TEST, JavaScopes.SYSTEM );

    /**
     * @parameter expression="${repositorySystemSession}"
     * @readonly
     */
    private RepositorySystemSession repositorySystemSession;

    /**
     * @component
     */
    private ProjectDependenciesResolver dependencyResolver;

    /**
     * @component
     */
    private MavenProjectHelper projectHelper;

    /**
     * @component
     */
    private Generator generator;

    /**
     * @component
     */
    private AdvicesParser advicesParser;

    @Override
    protected void doWithEclipse( final EclipseInstance eclipse )
        throws MojoExecutionException
    {
        Utils.failIfTouchpointClassesAvailableInClasspath();
        try
        {
            final DefaultRepositorySystemSession aetherSession =
                new DefaultRepositorySystemSession( repositorySystemSession );
            final ScopeDependencyGraphTransformer scopeTransformer =
                new ScopeDependencyGraphTransformer( includeScopes, excludeScopes );
            if ( aetherSession.getDependencyGraphTransformer() != null )
            {
                aetherSession.setDependencyGraphTransformer( new ChainedDependencyGraphTransformer(
                    aetherSession.getDependencyGraphTransformer(), scopeTransformer ) );
            }
            else
            {
                aetherSession.setDependencyGraphTransformer( scopeTransformer );
            }
            if ( !useTransitiveDependencies )
            {
                aetherSession.setDependencyGraphTransformer( new ChainedDependencyGraphTransformer(
                    aetherSession.getDependencyGraphTransformer(), new DropTransitiveGraphTransformer() ) );
            }

            final DefaultDependencyResolutionRequest resolution =
                new DefaultDependencyResolutionRequest( project, aetherSession );
            final DependencyResolutionResult resolutionResult = dependencyResolver.resolve( resolution );

            final DependencyNode dependencyTree = fixRootNode( resolutionResult.getDependencyGraph() );

            final Collection<Advice> advices = new ArrayList<Advice>();
            advices.addAll( extractAdvicesFromProject() );
            advices.addAll( extractAdvicesFromArtifacts() );

            final Collection<Artifact> metas = getMetaRequirements();

            final InstallableUnitsGeneratorAdaptor iuGenerator =
                new InstallableUnitsGeneratorAdaptor( eclipse.getService( BundlePublisher.class ) );

            final GeneratorResult result =
                generator.run( id, dependencyTree, advices,
                    versionPolicy == null ? new VersionPolicy() : versionPolicy, metas, project.getProperties(),
                    iuGenerator );

            FileUtils.deleteDirectory( updateSiteDirectory );

            {
                Map<String, String> properties = null;
                if ( tags != null )
                {
                    properties = new HashMap<String, String>();
                    properties.put( MetadataRepository.TAGS_PROPERTY, tags );
                }

                final MetadataRepository metadataRepositoryIO = eclipse.getService( MetadataRepository.class );
                metadataRepositoryIO.write( updateSiteDirectory.toURI(), result.getInstallableUnits(),
                    "sisu-generated-metadata", properties );
            }

            {
                Map<String, String> properties = null;
                if ( tags != null )
                {
                    properties = new HashMap<String, String>();
                    properties.put( ArtifactRepository.TAGS_PROPERTY, tags );
                }

                final ArtifactRepository artifactRepository = eclipse.getService( ArtifactRepository.class );
                artifactRepository.write( updateSiteDirectory.toURI(), result.getInstallableArtifacts(),
                    "sisu-generated-artifacts", properties );

                if ( resolveUpdateSiteArtifacts )
                {
                    getLog().debug( String.format( "Resolving artifacts from update site %s", updateSiteDirectory ) );
                    artifactRepository.resolve( updateSiteDirectory.toURI(), this );
                }
            }

            if ( attachUpdateSiteXmls )
            {
                attachUpdateSiteXmls();
            }
        }
        catch ( final DependencyResolutionException exception )
        {
            throw new MojoExecutionException( "Cannot build project dependency tree", exception );
        }
        catch ( final Exception e )
        {
            throw new MojoExecutionException( "Cannot generate update site", e );
        }
    }

    private Collection<Artifact> getMetaRequirements()
        throws ArtifactResolutionException
    {
        final Set<Artifact> metas = new HashSet<Artifact>();
        if ( useDefaultMetaRequirements )
        {
            metas.add( resolveArtifact( getPluginArtifact( "org.sonatype.p2.touchpoints", "org.sonatype.p2.touchpoint",
                "jar" ) ) );
            metas.add( resolveArtifact( getPluginArtifact( "org.sonatype.p2.touchpoints",
                "org.sonatype.p2.touchpoint.mixin.generic", "jar" ) ) );
            metas.add( resolveArtifact( getPluginArtifact( "org.sonatype.p2.touchpoints",
                "org.sonatype.p2.touchpoint.mixin.jsw", "jar" ) ) );
        }
        if ( metaRequirements != null )
        {
            for ( final MetaRequirement meta : metaRequirements )
            {
                metas.add( resolveArtifact( meta.groupId, meta.artifactId, meta.version, meta.type, meta.classifier ) );
            }
        }
        return metas;
    }

    private Collection<Advice> extractAdvicesFromProject()
        throws IOException
    {
        if ( !advicesDirectory.exists() )
        {
            return Collections.emptyList();
        }

        final DirectoryScanner ds = new DirectoryScanner();
        ds.setBasedir( advicesDirectory );
        ds.setIncludes( new String[] { "*.xml" } );
        ds.addDefaultExcludes();
        ds.scan();

        final Collection<Advice> advices = new ArrayList<Advice>();
        for ( final String adviceFile : ds.getIncludedFiles() )
        {
            InputStream in = null;
            try
            {
                in = new BufferedInputStream( new FileInputStream( new File( advicesDirectory, adviceFile ) ) );
                final Collection<Advice> parsed = advicesParser.parse( in, projectId );
                advices.addAll( parsed );
            }
            finally
            {
                IOUtil.close( in );
            }

        }
        return advices;
    }

    private Collection<Advice> extractAdvicesFromArtifacts()
        throws IOException, MojoExecutionException, ArtifactResolutionException
    {
        if ( adviceArtifacts == null || adviceArtifacts.isEmpty() )
        {
            return Collections.emptyList();
        }

        final Collection<Advice> advices = new ArrayList<Advice>();
        for ( final AdviceArtifact adviceArtifact : adviceArtifacts )
        {
            InputStream in = null;
            try
            {
                resolveVersionFromDependencyManagement( adviceArtifact );
                in = new BufferedInputStream( new FileInputStream( resolveArtifactFile( adviceArtifact ) ) );
                final Collection<Advice> parsed = advicesParser.parse( in, projectId );
                advices.addAll( parsed );
            }
            finally
            {
                IOUtil.close( in );
            }

        }
        return advices;
    }

    // this changes the root node to be the project not a plain dumb node
    private DependencyNode fixRootNode( final DependencyNode root )
        throws MojoExecutionException
    {
        final DefaultDependencyNode newRoot = new DefaultDependencyNode( root );
        newRoot.getChildren().addAll( root.getChildren() );

        Artifact artifact = null;
        if ( alternativeArtifact == null )
        {
            artifact = RepositoryUtils.toArtifact( project.getArtifact() );
        }
        else
        {
            final List<org.apache.maven.artifact.Artifact> attachedArtifacts = project.getAttachedArtifacts();
            for ( final org.apache.maven.artifact.Artifact attached : attachedArtifacts )
            {
                if ( ( attached.getType() == null ? alternativeArtifact.getType() == null : attached.getType().equals(
                    alternativeArtifact.getType() ) )
                    &&
                    ( attached.getClassifier() == null ? alternativeArtifact.getClassifier() == null
                                    : attached.getClassifier().equals( alternativeArtifact.getClassifier() ) ) )
                {
                    artifact = RepositoryUtils.toArtifact( attached );
                    break;
                }
            }
        }
        newRoot.setDependency( new Dependency( artifact, "compile" ) );

        return newRoot;
    }

    private void attachUpdateSiteXmls()
    {
        projectHelper.attachArtifact( project, "xml", "p2Artifacts", new File( updateSiteDirectory, "artifacts.xml" ) );
        projectHelper.attachArtifact( project, "xml", "p2Content", new File( updateSiteDirectory, "content.xml" ) );
    }

    private static class InstallableUnitsGeneratorAdaptor
        implements InstallableUnitsGenerator
    {
        private final BundlePublisher bundlePublisher;

        InstallableUnitsGeneratorAdaptor( final BundlePublisher bundlePublisher )
        {
            this.bundlePublisher = bundlePublisher;

        }

        public Collection<InstallableUnit> generate( final GenerateSettings settings, final File... bundles )
        {
            final boolean generateProvided = isOn( settings.getP2iu() ) || isOn( settings.getP2provided() );
            final boolean generateRequirements = isOn( settings.getP2iu() ) || isOn( settings.getP2required() );
            final boolean generateManifest = isOn( settings.getP2iu() ) || isOn( settings.getP2manifest() );

            return org.sonatype.sisu.assembler.internal.Utils.adaptFromBridgeModel( bundlePublisher.generate(
                generateProvided, generateRequirements,
                generateManifest,
                bundles ) );
        }

        private static boolean isOn( final OnOffSwitch onOffSwitch )
        {
            return onOffSwitch != null && onOffSwitch.isOn();
        }
    }

}
