/*
 * Copyright (c) 2007-2010 Sonatype, Inc.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/epl-v10.html
 */
package org.sonatype.plugins.sisu.assembler;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.File;
import java.io.FileReader;
import java.io.FileWriter;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.Map;

public class PermissionRegistry
{
    private final Map<File, Entry> entries;

    private final File permissionsFile;

    /**
     * True, if properties have been loaded from properties file. Used for lazy loading.
     */
    private boolean loaded;

    public PermissionRegistry()
    {
        this( null );
    }

    public PermissionRegistry( final File permissionsFile )
    {
        entries = new HashMap<File, Entry>();
        this.permissionsFile = permissionsFile;
        loaded = false;
        load();
    }

    public void reset()
    {
        entries.clear();
        loaded = false;

        load();
    }

    private void load()
    {
        if ( loaded || permissionsFile == null || !permissionsFile.exists() )
        {
            return;
        }

        BufferedReader reader = null;
        try
        {
            reader = new BufferedReader( new FileReader( permissionsFile ) );
            String line = null;
            while ( ( line = reader.readLine() ) != null )
            {
                final int startOfPermissions = line.lastIndexOf( ',' );
                if ( startOfPermissions > 0 )
                {
                    final File target = new File( line.substring( 0, startOfPermissions ) );
                    final String permissions = line.substring( startOfPermissions + 1 );
                    entries.put( target, new Entry( target, permissions ) );
                }
            }
        }
        catch ( final IOException e )
        {
            throw new RuntimeException( e );
        }
        finally
        {
            if ( reader != null )
            {
                try
                {
                    reader.close();
                }
                catch ( final IOException ignore )
                {
                    // ignore
                }
            }
        }
        loaded = true;
    }

    public void save()
    {
        if ( permissionsFile == null )
        {
            return;
        }
        BufferedWriter writter = null;
        try
        {
            writter = new BufferedWriter( new FileWriter( permissionsFile ) );
            for ( final Entry entry : entries.values() )
            {
                writter.write( entry.getTarget().getPath() );
                writter.write( "," );
                writter.write( entry.getPermission() );
                writter.newLine();
            }
        }
        catch ( final IOException e )
        {
            throw new RuntimeException( e );
        }
        finally
        {
            if ( writter != null )
            {
                try
                {
                    writter.close();
                }
                catch ( final IOException e )
                {
                    throw new RuntimeException( e );
                }
            }
        }
    }

    public void setPermissions( final File target, final String permission )
    {
        if ( permission != null )
        {
            final Entry entry = new Entry( target, permission );
            entries.put( target, entry );
        }
        else
        {
            entries.remove( target );
        }
    }

    public String getPermissions( final File target )
    {
        final Entry entry = entries.get( target );

        if ( entry != null )
        {
            return entry.getPermission();
        }

        return null;
    }

    public Collection<Entry> getPermissions()
    {
        return Collections.unmodifiableCollection( entries.values() );
    }

    public class Entry
    {
        private final File target;

        private final String permission;

        public Entry( final File target, final String permission )
        {
            this.target = target;
            this.permission = permission;
        }

        public File getTarget()
        {
            return target;
        }

        public String getPermission()
        {
            return permission;
        }

    }
}
